# This software was produced by NIST, an agency of the U.S. government,
# and by statute is not subject to copyright in the United States.
# Recipients of this software assume all responsibilities associated
# with its operation, modification and maintenance. However, to
# facilitate maintenance we ask that before distributing modified
# versions of this software, you first contact the authors at
# oof_manager@ctcms.nist.gov.

'''
This module contains the TridiagMatrix class to represent tridiagonal
matrices and associated routines for linear algebra, such as a direct
solver based on LU decomposition (Thomas algorithm) with linear time
complexity.
'''


from numpy import *


ZERO = 10**-15


class TridiagMatrix:
    '''
    The TridiagMatrix class to store tridiagonal matrices of the form

       ( b(0) c(0)             a(0) )
       ( a(1) b(1) c(1)             )
       (      a(2)  .  .            )
       (            .  .  .         )
       (               .  .  c(n-1) )
       ( c(n)            a(n) b(n)  )

    It is used for implementing efficient linear algebra for
    tridiagonal matrices.

    The following methods are available:
       setDiag( d, bandNo )
       getDiag( d, bandNo )
       matvec( v )
       size( )

    '''

    def __init__(self,N_or_a,b=None,c=None):

        if (b == None) or (c == None):
            self.N  = N_or__aa
            self.__a = zeros(N)
            self.__b = zeros(N)
            self.__c = zeros(N)

        elif (len(b) == len(N_or_a)) and (len(c) == len(N_or_a)):
            self.N  = len(N_or_a)
            self.__a = N_or_a
            self.__b = b
            self.__c = c


    def size(self):
        'Tridiag.size() returns the size of the tridiagonal matrix'
        return self.N


    def setDiag(self,d,bandNo):
        '''Tridiag.setDiag( d, bandNo ) sets the band specified with bandNo
           to the values given in d. The parameter bandNo should be one of
           -1,0,1.'''

        if not (bandNo in [-1,0,1]):
            raise Exception, 'Parameter bandNo should be one of -1,0,1'

        elif len(d) != self.N:
            raise Exception, 'Size of input vector d should be %d' % self.N

        elif bandNo == -1:
            self.__a = d.copy()

        elif bandNo == 0:
            self.__b = d.copy()

        elif bandNo == 1:
            self.__c = d.copy()


    def getDiag(self,bandNo):
        '''Tridiag.getDiag( d, bandNo ) returns the values in the band
           specified with bandNo, which should be one of -1,0,1.'''

        if not (bandNo in [-1,0,1]):
            raise Exception, 'Parameter bandNo should be one of -1,0,1'

        elif bandNo == -1:
            return self.__a

        elif bandNo == 0:
            return self.__b

        elif bandNo == 1:
            return self.__c


    def matvec(self,v):
        '''Tridiag.matvec( v ) returns the matrix-vector product with
           the vector v.'''

        N = len(v)

        if N != self.N: raise Exception, 'The matrix and the vector should be of the same dimension!'

        a = self.__a;   b = self.__b;   c = self.__c

        u = b * v
        u[0]     += a[0] * v[N-1]
        u[1:N]   += a[1:N] * v[0:N-1]
        u[N-1]   += c[N-1] * v[0]
        u[0:N-1] += c[0:N-1] * v[1:N]

        return u



def tridiagSolve(A,d):
    '''tridiagSolve( A, d ) takes a tridiagonal matrix A and right
       hand side vector d and returns the solution x of the linear
       system   A x = d.'''


    a = A.getDiag(-1)
    b = A.getDiag(0)
    c = A.getDiag(1)
    N = A.size()

    if len(d) != N:
        raise Exception, 'Matrix A and right hand side vector d should be the same size'


    periodicTridiagSystem = (abs(a[0]) > ZERO) or (abs(c[N-1]) > ZERO)


    if periodicTridiagSystem:
        b[0]   -= a[0]
        b[N-1] -= c[N-1]


    l,m = tridiagLU( a[1:], b, c[:N-1] )

    y = tridiagSolveFromLU( l, m, c[:N-1], d )


    if periodicTridiagSystem:

        u = zeros(N);  v = zeros(N)
        u[0] = a[0];   u[N-1] = c[N-1]
        v[0] = 1;      v[N-1] = 1

        z = tridiagSolveFromLU( l, m, c[:N-1], u )

        x = y - z * dot(v,y) / (1 + dot(v,z))

        b[0]   += a[0]
        b[N-1] += c[N-1]

    else:
        x = y

    return x



def tridiagLU(a,b,c):
    '''tridiagLU( a, b, c ) computes the LU decomposition of the tridiagonal matrix

             ( b(0) c(0)             a(0) )
             ( a(1) b(1) c(1)             )
         A = (      a(2)  .  .            )
             (            .  .  .         )
             (               .  .  c(n-1) )
             ( c(n)            a(n) b(n)  )

       from the given three bands a,b,c and returns two vectors l,m that give
       the LU decomposition
           A = L U
       where
           L = Id + diag(-1,l),   U = diag(0,m) + diag(1,r)
    '''

    N = len(b);   m = zeros(N);   l = zeros(N);

    m[0] = b[0]
    for i in range(N-1):
        l[i] = a[i] / m[i]
        m[i+1] = b[i+1] - l[i]*c[i]

    return (l,m)



def tridiagSolveFromLU(l,m,c,d):
    '''tridiagSolveFromLU( l, m, c, d ) takes the LU decomposition l,m
       (which should be calculated by tridiagLU), the third band c
       (bandNo = 1), and right hand side d of the tridiagonal linear
       system and returns its solution. See documentation of tridiagLU
       for more informations on the arguments.
    '''

    N = len(d)

    x = zeros(N);  y = zeros(N);

    y[0] = d[0]
    for i in range(1,N):
        y[i] = d[i] - l[i-1]*y[i-1]

    x[N-1] = y[N-1] / m[N-1]
    for i in range(N-2,-1,-1):
        x[i] = (y[i] - c[i]*x[i+1]) / m[i]

    return x
